"""This module contains the meta for the default load profile
simulator.

All profiles are normalized to an annual consumption of 1000 kWh/a.

"""


MODELS = [
    "H0",  # Households
    "G0",  # Commercial (general)
    "G1",  # Commercial (weekday 8-18)
    "G2",  # Commercial (evening)
    "G3",  # Commercial (continuous)
    "G4",  # Commercial (Shop/Barber)
    "G5",  # Commercial (Bakery)
    "G6",  # Commercial (weekend)
    "L0",  # Agriculture (general)
    "L1",  # Acriculture (milk/animal breeding)
    "L2",  # Agriculture (other)
]

CONFIG = {
    "public": True,
    "params": [
        "p_mwh_per_a",
        "interpolate",
        "randomize_data",
        "randomize_cos_phi",
    ],
    "attrs": ["p_mw", "q_mvar", "cos_phi"],
}

META = {
    "type": "time-based",
    "models": {model: CONFIG for model in MODELS},
    "extra_methods": ["get_data_info"],
}



""" This module contains the DLP simulator."""
import os
from datetime import datetime

import h5py
import mosaik_api
import numpy as np
from midas.util.base_data_simulator import BaseDataSimulator
from midas.util.dateformat import GER
from midas.util.runtime_config import RuntimeConfig

from . import LOG
from .meta import META
from .model import DLPModel


class DLPSimulator(BaseDataSimulator):
    """The DLP simulator."""

    def __init__(self):
        super().__init__(META)
        self.sid = None
        self.models = dict()
        self.num_models = dict()
        self.data = dict()

    def init(self, sid, **sim_params):
        """Called exactly ones after the simulator has been started.

        Parameters
        ----------
        sid : str
            Simulator ID for this simulator.
        step_size : int, optional
            Step size for this simulator. Defaults to 900.

        Returns
        -------
        dict
            The meta dict (set by *mosaik_api.Simulator*).
        """
        super().init(sid, **sim_params)

        # We want to use the provided timezone
        self.now_dt = datetime.strptime(sim_params["start_date"], GER)

        data_path = sim_params.get(
            "data_path", RuntimeConfig().paths["data_path"]
        )
        file_path = os.path.join(
            data_path, sim_params.get("filename", "DefaultLoadProfiles.hdf5")
        )
        LOG.debug("Using db file at %s.", file_path)

        with h5py.File(file_path, "r") as h5f:
            for profile in h5f:
                self.data.setdefault(profile, dict())
                for season in h5f[profile]:
                    self.data[profile].setdefault(season, dict())
                    for day in h5f[profile][season]:
                        self.data[profile][season][day] = np.array(
                            list(h5f[profile][season][day])
                        )
        return self.meta

    def create(self, num, model, **model_params):
        """Initialize the simulation model instance (entity).

        Parameters
        ----------
        num : int
            The number of models to create.
        model : str
            The name of the models to create. Must be present inside
            the simulator's meta.

        Returns
        -------
        list
            A list with information on the created entity.
        """

        entities = list()
        self.num_models.setdefault(model, 0)
        for _ in range(num):
            eid = f"{model}-{self.num_models[model]}"
            profile = model
            self.models[eid] = DLPModel(
                data=self.data[profile],
                p_mwh_per_a=model_params["p_mwh_per_a"],
                seed=self.rng.randint(self.seed_max),
                interpolate=model_params.get("interpolate", self.interpolate),
                randomize_data=model_params.get(
                    "randomize_data", self.randomize_data
                ),
                randomize_cos_phi=model_params.get(
                    "randomize_cos_phi", self.randomize_cos_phi
                ),
            )

            self.num_models[model] += 1
            entities.append({"eid": eid, "type": model})

        return entities

    def step(self, time, inputs, max_advance=0):
        """Perform a simulation step.

        Parameters
        ----------
        time : int
            The current simulation step (by convention in seconds since
            simulation start.
        inputs : dict
            A *dict* containing inputs for entities of this simulator.

        Returns
        -------
        int
            The next step this simulator wants to be stepped.

        """
        LOG.debug("At step %d received inputs %s", time, inputs)

        return super().step(time, inputs)

    def get_data(self, outputs):
        """Return the requested output (if feasible).

        Parameters
        ----------
        outputs : dict
            A *dict* containing requested outputs of each entity.

        Returns
        -------
        dict
            A *dict* containing the values of the requested outputs.

        """

        data = super().get_data(outputs)

        LOG.debug("Gathered outputs %s", data)

        return data

    def get_data_info(self):
        info = {
            key: {"p_mwh_per_a": model.p_mwh_per_a}
            for key, model in self.models.items()
        }
        for model in self.meta["models"]:
            info[f"num_{model.lower()}"] = self.num_models.get(model, 0)

        return info


if __name__ == "__main__":
    mosaik_api.start_simulation(DLPSimulator())



"""This module contains the default load profile data model."""
import numpy as np

from midas.util.compute_q import compute_q


class DLPModel:
    """A model to handle a default load profile."""

    def __init__(self, data, p_mwh_per_a, **params):
        self.data = data
        self.p_mwh_per_a = p_mwh_per_a

        self._rng = np.random.RandomState(params.get("seed", None))

        self._interpolate = params.get("interpolate", False)
        self._randomize_data = params.get("randomize_data", False)
        self._randomize_cos_phi = params.get("randomize_cos_phi", False)
        self._randomize_data_scale = params.get("randomize_data_scale", 0.05)
        self._randomize_cos_phi_scale = params.get(
            "randomize_cos_phi_scale", 0.01
        )

        # Inputs
        self.now_dt = None
        self.cos_phi = None

        # Outputs
        self.p_mw = None
        self.q_mvar = None

    def step(self):
        """Generate the data for the next time interval defined by
        now_dt.

        """
        if self.now_dt.month in {1, 2, 12}:
            season = "winter"
        elif self.now_dt.month in {6, 7, 8}:
            season = "summer"
        else:
            season = "transition"

        if self.now_dt.weekday() == 5:
            day = "saturday"
        elif self.now_dt.weekday() == 6:
            day = "sunday"
        else:
            day = "weekday"

        # +1 because now dt is in UTC and DLP start in UTC+1
        idx = int(self.now_dt.hour * 4 + self.now_dt.minute // 15)
        idx_interp = self.now_dt.hour * 4 + self.now_dt.minute / 15
        if self._interpolate and idx != idx_interp:
            cur_sec = self.now_dt.minute * 60 + self.now_dt.second
            f_p = self.data[season][day][idx : idx + 2]
            if len(f_p) == 1:
                f_p = np.array([f_p[0], self.data[season][day][0]])
            self.p_mw = np.interp(cur_sec, [0, 900], f_p)
        else:
            self.p_mw = self.data[season][day][idx]

        if self._randomize_data:
            self.p_mw *= self._rng.normal(
                scale=self._randomize_data_scale, loc=1.0
            )
            self.p_mw = max(0, self.p_mw)

        if self._randomize_cos_phi:
            self.cos_phi = max(
                0,
                min(
                    2 * np.pi,
                    self._rng.normal(
                        scale=self._randomize_cos_phi_scale, loc=0.9
                    ),
                ),
            )

        self.p_mw *= self.p_mwh_per_a * 1e-6
        self.q_mvar = compute_q(self.p_mw, self.cos_phi)
